#include "stm32f0xx_conf.h"

void USART2_SendString(const char *str);
void USART2_IRQHandler(void);
void USART2_SendHexByte(uint8_t c);
void HandleMacCommand();
void SendMacResponse(uint8_t respbyte);
uint16_t PollPS2Frame(uint32_t clock, uint32_t data);
void SendPS2Command(uint8_t cmd, uint32_t clock, uint32_t data);
void SetPS2LEDState(uint8_t cmd);
void SetPS2Repeat(void);
uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended);
void HandlePS2Keyboard(void);
void HandlePS2Mouse(void);
void SetupPS2Mouse(void);

#define kBuffSize  32
static __IO unsigned long tenustimer;
static __IO uint8_t mackey[kBuffSize];
static __IO uint8_t keyindex;
static __IO uint8_t key2index;
static __IO uint8_t keypad;

static __IO uint16_t mousex;
static __IO uint16_t mousey;
static __IO uint16_t mousebits;

#define kPS2KbdClockPin             GPIO_Pin_9
#define kPS2KbdDataPin              GPIO_Pin_8
#define kMacKbdClockPin             GPIO_Pin_7
#define kMacKbdDataPin              GPIO_Pin_6

#define kPS2MouseClockPin           GPIO_Pin_1
#define kPS2MouseDataPin            GPIO_Pin_0

#define kMSEX1 GPIO_Pin_3
#define kMSEX2 GPIO_Pin_2
#define kMSEY1 GPIO_Pin_5
#define kMSEY2 GPIO_Pin_4
#define kMSESW GPIO_Pin_7

#define kMacKbdStateIdle            0
#define kMacKbdStateReadSettle      1
#define kMacKbdStateReading         2
#define kMacKbdStateWriteSettle     3
#define kMacKbdStateWriting         4
#define kMacKbdStateInquiryWait     5
#define kMacKbdStateTest            6
#define kMacKbdStateInstantWait     7

struct mackbd {
	uint8_t state;
	uint8_t bits;
	uint8_t clockstate;
	uint32_t delay;
	uint8_t cmd;
	uint8_t response;
};
static __IO struct mackbd mackbd;

void handle_mouseaxis(__IO uint16_t *movement, uint32_t pinone, uint32_t pintwo) {
	static uint8_t counter = 100;

	if(*movement & 0x00FF) {
		uint8_t movecount = *movement & 0x00FF;

		if(counter--) return;
		counter = 100;

		*movement = (*movement & 0xFF00) | (movecount-1);
		if((*movement & 0x00FF) < 2) return;

		if(!(*movement & 0x0100)) {
			if(mousebits & pinone) {
				GPIO_ResetBits(GPIOC, pintwo);
				GPIO_ResetBits(GPIOC, pinone);
				mousebits &= ~(pinone | pintwo);
			} else {
				GPIO_SetBits(GPIOC, pintwo);
				GPIO_SetBits(GPIOC, pinone);
				mousebits |= pinone | pintwo;
			}
		} else {
			if(mousebits & pinone) {
				GPIO_SetBits(GPIOC, pintwo);
				mousebits |= pintwo;
				GPIO_ResetBits(GPIOC, pinone);
				mousebits &= ~pinone;
			} else {
				GPIO_ResetBits(GPIOC, pintwo);
				mousebits &= ~pintwo;
				GPIO_SetBits(GPIOC, pinone);
				mousebits |= pinone;
			}
		}
	}
	return;
}

void SysTick_Handler(void) {
	GPIO_InitTypeDef GPIO_InitStructure;
	static uint8_t respbit;
	tenustimer++;

	handle_mouseaxis(&mousex, kMSEX1, kMSEX2);
	handle_mouseaxis(&mousey, kMSEY1, kMSEY2);
	switch(mackbd.state) {
	case kMacKbdStateIdle:
		if(GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) == 0) {
			mackbd.state = kMacKbdStateReadSettle;
			mackbd.delay = 18;
			mackbd.bits = 0;
			mackbd.cmd = 0;
			mackbd.clockstate = 3;
		}
		break;
	case kMacKbdStateReadSettle:
		if(--mackbd.delay == 0) {
			mackbd.state = kMacKbdStateReading;
		}
		break;
	case kMacKbdStateReading:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.delay = 8;
				mackbd.clockstate = 4;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					// Done clocking in all the bits
					mackbd.state = kMacKbdStateIdle;
					HandleMacCommand();
				} else {
					// Not done, transition to next clock
					mackbd.delay = 18;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
				}
			}
			break;
		case 3:
			// We just entered the read state, bring the clock low
			mackbd.clockstate = 0;
			mackbd.delay = 18;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			break;
		case 4: /* read the data 80uS into the rising clock */
			if(--mackbd.delay == 0) {
				mackbd.cmd |= GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) << (7 - mackbd.bits++);
				mackbd.delay = 14;
				mackbd.clockstate = 1;
			}
			break;
		default:
			break;
		};
		break;
	case kMacKbdStateWriteSettle:
		if(--mackbd.delay == 0) {
			if(GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) == 0) {
				/* "... the computer then indicates that it is ready 
				 * to receive the keyboard's response by setting the 
				 * Keyboard Data line high
				 */
				mackbd.delay = 1;
			} else {
				mackbd.state = kMacKbdStateWriting;
				mackbd.clockstate = 3;
				mackbd.bits = 0;
			}
		}
		break;
	case kMacKbdStateWriting:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.delay = 18;
				mackbd.clockstate = 1;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					mackbd.state = kMacKbdStateIdle;
					GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
					GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
					GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
					GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
					GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
					GPIO_Init(GPIOC, &GPIO_InitStructure);
				} else {
					respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
					mackbd.delay = 16;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
					respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
					mackbd.bits++;
				}
			}
			break;
		case 3:
			GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
			GPIO_Init(GPIOC, &GPIO_InitStructure);

			respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
			mackbd.bits++;
			mackbd.clockstate = 0;
			mackbd.delay = 16;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
			break;
			
		default:
			break;
		};
		break;
	case kMacKbdStateInstantWait:
	case kMacKbdStateInquiryWait:
		if(mackey[keyindex]) {
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = mackey[keyindex];
			mackey[keyindex] = 0;
			keyindex = (keyindex + 1) % kBuffSize;
		} else if(--mackbd.delay == 0) {
			// If no keyboard press, send a null
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = 0x7B;
		}
		break;
	case kMacKbdStateTest:
		mackbd.state = kMacKbdStateWriteSettle;
		mackbd.delay = 1;
		mackbd.response = 0x7D;
		break;
	default:
		break;
	};
}

int main(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	GPIO_InitTypeDef GPIO2_struct;
	USART_InitTypeDef USART_InitStructure;
	NVIC_InitTypeDef NVIC_InitStructure;
	int i;

	mousex = 0;
	mousey = 0;
	tenustimer = 0;
	keyindex = 0;
	key2index = 0;
	keypad = 0;
	for(i = 0; i < kBuffSize; i++) mackey[i] = 0;

	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);

	// Mouse

	GPIO2_struct.GPIO_Pin = kMSEX1 | kMSEX2 | kMSEY1 | kMSEY2;
	GPIO2_struct.GPIO_Mode = GPIO_Mode_OUT;
	GPIO2_struct.GPIO_OType = GPIO_OType_PP;
	GPIO2_struct.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO2_struct.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO2_struct);

	// Mac Keyboard
	// Clock
	GPIO_InitStructure.GPIO_Pin = kMacKbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Data
	GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// PS2 keyboard
	GPIO_InitStructure.GPIO_Pin = kPS2KbdDataPin | kPS2KbdClockPin | kPS2MouseDataPin | kPS2MouseClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);



	// UART2
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2,ENABLE);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource2, GPIO_AF_1);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource3, GPIO_AF_1);
	GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_2 | GPIO_Pin_3;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &GPIO_InitStructure);


	NVIC_InitStructure.NVIC_IRQChannel = USART2_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPriority = 0;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

	USART_InitStructure.USART_BaudRate = 9600;
	USART_InitStructure.USART_WordLength = USART_WordLength_8b;
	USART_InitStructure.USART_StopBits = USART_StopBits_1;
	USART_InitStructure.USART_Parity = USART_Parity_No;
	USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USART_Init(USART2, &USART_InitStructure);
	USART_Cmd(USART2,ENABLE);
	USART_ITConfig(USART2, USART_IT_RXNE, ENABLE);

	// Mouse button
	GPIO_InitStructure.GPIO_Pin = kMSESW;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	USART2_SendString("Starting up\n\r");
	GPIO_SetBits(GPIOC, kMacKbdClockPin);

	mousebits = 0;
	GPIO_SetBits(GPIOC, kMSEX1 | kMSEY1);
	mousebits |= (kMSEX1 | kMSEY1);
	GPIO_ResetBits(GPIOC, kMSEX2 | kMSEY2);
	mousebits &= ~(kMSEX2 | kMSEY2);

	GPIO_ResetBits(GPIOA, kMSESW);

	mackbd.state = kMacKbdStateIdle;

	// set a ~10us timer
	SysTick_Config(SystemCoreClock/100000);
	USART2_SendString("Timer configured\n\r");

	SetPS2Repeat();
	SetupPS2Mouse();

	while(1) {
		if(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0) {
			HandlePS2Keyboard();
		}

		if(GPIO_ReadInputDataBit(GPIOC, kPS2MouseDataPin) == 0) {
			HandlePS2Mouse();
		}
	}

}

void HandlePS2Mouse(void)
{
	uint16_t ps2frame = 0;
	uint8_t status;
	uint8_t xmove;
	uint8_t ymove;
	uint8_t absx;
	uint8_t absy;

	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	status = (ps2frame >> 1) & 0xFF;
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	xmove = (ps2frame >> 1) & 0xFF;
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ymove = (ps2frame >> 1) & 0xFF;

	absx = xmove;
	absy = ymove;
	//xmove += 1;
	//ymove += 1;

	//xmove &= 0x7F;
	//ymove &= 0x7F;

	if(xmove && ((mousex & 0x00FF) == 0)) {
		mousex = 0;
		if(status & 0x10) {
			absx = ~xmove;
			mousex |= 0x0100;
		}
		mousex |= absx;
	}
	if(ymove && ((mousey & 0x00FF) == 0)) {
		mousey = 0;
		if(status & 0x20) {
			absy = ~ymove;
			mousey |= 0x0100;
		}
		mousey |= absy;
	}

	// Left button == mouse button
	if(status & 0x01) {
		GPIO_ResetBits(GPIOA, kMSESW);
	} else {
		GPIO_SetBits(GPIOA, kMSESW);
	}
#if 0
	USART2_SendHexByte((mousex >> 8) & 0x00FF);
	USART2_SendHexByte(mousex & 0x00FF);
	USART2_SendString(" ");
	USART2_SendHexByte((uint8_t)((mousey >> 8) & 0x00FF));
	USART2_SendHexByte((uint8_t)(mousey & 0x00FF));
	USART2_SendString("\n\r");
#endif

#if 0
	USART2_SendHexByte(status);
	USART2_SendString(" ");
	USART2_SendHexByte(xmove);
	USART2_SendString(" ");
	USART2_SendHexByte(ymove);
	USART2_SendString("\n\r");
#endif
}

void HandlePS2Keyboard(void) 
{
	static uint8_t ps2ledstate = 0;
	static uint8_t ps2lastcode = 0;
	uint16_t ps2frame = 0;
	uint8_t ps2scancode;
	uint8_t tmpcode = 0;

	ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
	ps2scancode = tmpcode = (ps2frame >> 1) & 0xFF;
	if(ps2scancode == 0xF0) {
		// Keyup, poll in the scancode of the key that has been released
		ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
		ps2scancode = ps2frame >> 1;
		// Ignore keyup capslock if we're turning it on
		if(((ps2scancode) == 0x58) && (ps2ledstate & 1<<2)) {
		} else {
			PS2ToMacScancode(ps2scancode, 1, 0);
		}
	} else if(ps2scancode == 0xE0) {
		// two byte scan code
		ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
		ps2scancode = ps2frame >> 1;
		if(ps2scancode == 0xF0) {
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			if(ps2scancode == 0x7C) {
				// prntscrn
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be E0
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be F0
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be 12
			} else {
				PS2ToMacScancode(ps2scancode, 1, 1);
			}
		} else if(ps2scancode == 0x12) {
			// prntscrn
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			// should be E0
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			// should be 7C
		} else {
			PS2ToMacScancode(ps2scancode, 0, 1);
		}
	} else {
		if(ps2scancode == ps2lastcode) {
			return;
		}
		// If it's capslock, turn on the capslock led
		if((ps2scancode) == 0x58) {
			if(ps2ledstate & (1<<2)) {
				ps2ledstate &= ~(1<<2);
			} else {
				ps2ledstate |= 1<<2;
			}
			SetPS2LEDState(ps2ledstate);
		}
		PS2ToMacScancode(ps2scancode, 0, 0);
	}
	//USART2_SendHexByte(ps2scancode);
	ps2lastcode = tmpcode;
	return;
}

uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended)
{
	uint8_t maccode = 0;
	keypad = 0;
	if(!extended) {
		switch(code) {
		case 0x1C: maccode = 0x01; break; // A
		case 0x32: maccode = 0x17; break; // B
		case 0x21: maccode = 0x11; break; // C
		case 0x23: maccode = 0x05; break; // D
		case 0x24: maccode = 0x1D; break; // E
		case 0x2B: maccode = 0x07; break; // F
		case 0x34: maccode = 0x0B; break; // G
		case 0x33: maccode = 0x09; break; // H
		case 0x43: maccode = 0x45; break; // I
		case 0x3B: maccode = 0x4D; break; // J
		case 0x42: maccode = 0x51; break; // K
		case 0x4B: maccode = 0x4B; break; // L  // woah
		case 0x3A: maccode = 0x5D; break; // M
		case 0x31: maccode = 0x5B; break; // N
		case 0x44: maccode = 0x3F; break; // O
		case 0x4D: maccode = 0x47; break; // P
		case 0x15: maccode = 0x19; break; // Q
		case 0x2D: maccode = 0x1F; break; // R
		case 0x1B: maccode = 0x03; break; // S
		case 0x2C: maccode = 0x23; break; // T
		case 0x3C: maccode = 0x41; break; // U
		case 0x2A: maccode = 0x13; break; // V
		case 0x1D: maccode = 0x1B; break; // W
		case 0x22: maccode = 0x0F; break; // X
		case 0x35: maccode = 0x21; break; // Y
		case 0x1A: maccode = 0x0D; break; // Z
		case 0x45: maccode = 0x3B; break; // 0
		case 0x16: maccode = 0x25; break; // 1
		case 0x1E: maccode = 0x27; break; // 2
		case 0x26: maccode = 0x29; break; // 3
		case 0x25: maccode = 0x2B; break; // 4
		case 0x2E: maccode = 0x2F; break; // 5
		case 0x36: maccode = 0x2D; break; // 6
		case 0x3D: maccode = 0x35; break; // 7
		case 0x3E: maccode = 0x39; break; // 8
		case 0x46: maccode = 0x33; break; // 9
		case 0x0E: maccode = 0x65; break; // `
		case 0x4E: maccode = 0x37; break; // -
		case 0x55: maccode = 0x31; break; // =
		case 0x5D: maccode = 0x55; break; // backslash
		case 0x66: maccode = 0x67; break; // bkspc
		case 0x29: maccode = 0x63; break; // space
		case 0x0D: maccode = 0x61; break; // tab
		case 0x58: maccode = 0x73; break; // caps
		case 0x12: maccode = 0x71; break; // left shift
		case 0x59: maccode = 0x71; break; // right shift
		case 0x11: maccode = 0x75; break; // left alt = opt
		case 0x5A: maccode = 0x49; break; // enter
		case 0x41: maccode = 0x57; break; // ,
		case 0x49: maccode = 0x5F; break; // .
		case 0x4A: maccode = 0x59; break; // /
		case 0x14: maccode = 0x77; break; // L control
		case 0x76: maccode = 0x6B; break; // ESC
		case 0x52: maccode = 0x4F; break; // '
		case 0x7C: maccode = 0x79; keypad = 0x05; break; // keypad *
		case 0x7B: maccode = 0x79; keypad = 0x1D; break; // keypad -
		case 0x79: maccode = 0x79; keypad = 0x0D; break; // keypad +
		case 0x71: maccode = 0x79; keypad = 0x03; break; // keypad .
		case 0x70: maccode = 0x79; keypad = 0x25; break; // keypad 0
		case 0x69: maccode = 0x79; keypad = 0x27; break; // keypad 1
		case 0x72: maccode = 0x79; keypad = 0x29; break; // keypad 2
		case 0x7A: maccode = 0x79; keypad = 0x2B; break; // keypad 3
		case 0x6B: maccode = 0x79; keypad = 0x2D; break; // keypad 4
		case 0x73: maccode = 0x79; keypad = 0x2F; break; // keypad 5
		case 0x74: maccode = 0x79; keypad = 0x31; break; // keypad 6
		case 0x6C: maccode = 0x79; keypad = 0x33; break; // keypad 7
		case 0x75: maccode = 0x79; keypad = 0x37; break; // keypad 8
		case 0x7D: maccode = 0x79; keypad = 0x39; break; // keypad 9
		default: break;
		}
	} else {
		switch(code) {
		case 0x1F: maccode = 0x6F; break; // left windows key = cmd
		case 0x27: maccode = 0x6F; break; // right windows key = cmd
		case 0x11: maccode = 0x75; break; // right alt = cmd
		case 0x5A: maccode = 0x79; keypad = 0x19; break; // keypad enter
		case 0x4A: maccode = 0x79; keypad = 0x1B; break;  // keypad /
		case 0x75: maccode = 0x79; keypad = 0x1B; break;  // up -> keypad up
		case 0x6B: maccode = 0x79; keypad = 0x0D; break;  // left arrow
		case 0x72: maccode = 0x79; keypad = 0x11; break;  // down arrow
		case 0x74: maccode = 0x79; keypad = 0x05; break;  // right arrow
		case 0x14: maccode = 0x77; break; // R control
		}
	}

	if(keyup && maccode) {
		if(keypad) {
			keypad |= 0x80;
		} else {
			maccode |= 0x80;
		}
	}

	if(maccode == 0) maccode = 0x7B;

	__disable_irq();
	if(((key2index+1)%kBuffSize) != keyindex) {
		mackey[key2index] = maccode;
		key2index = (key2index+1) % kBuffSize;
		if(keypad) {
			mackey[key2index] = keypad;
			key2index = (key2index+1) % kBuffSize;
		}
	}
	__enable_irq();

	return maccode;
}

void SetupPS2Mouse(void)
{
	uint16_t ps2frame = 0;
	uint8_t ps2scancode;
	uint8_t resetresult;
	uint8_t mouseid;
	unsigned long tmp1 = tenustimer;

	while(tenustimer < (tmp1+10));
	
	USART2_SendString("Sending mouse reset\n\r");
	// reset
	SendPS2Command(0xFF, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	USART2_SendString("Got: ");
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");
	
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	resetresult = ps2frame >> 1;

	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	mouseid = ps2frame >> 1;

	USART2_SendString("Reset results: ");
	USART2_SendHexByte(resetresult);
	USART2_SendString(" ");
	USART2_SendHexByte(mouseid);
	USART2_SendString("\n\r");

	// Set sample rate
	SendPS2Command(0xF3, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x0A, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack

	// Read device type
	SendPS2Command(0xF2, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be mouseid

	// Set resolution
	SendPS2Command(0xE8, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x03, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack

	// Set scaling
	SendPS2Command(0xE6, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack

	// Set sample rate
	SendPS2Command(0xE6, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x28, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack

	// Enable
	SendPS2Command(0xF4, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
}

void SetPS2LEDState(uint8_t cmd)
{
	unsigned long tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(0xED, kPS2KbdClockPin, kPS2KbdDataPin);
	// wait for ack 0xFA
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
	tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(cmd, kPS2KbdClockPin, kPS2KbdDataPin);
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
}

void SetPS2Repeat(void)
{
	unsigned long tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(0xF8, kPS2KbdClockPin, kPS2KbdDataPin);
	// wait for ack 0xFA
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
}



void SendPS2Command(uint8_t cmd, uint32_t clock, uint32_t data)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	unsigned long tmp1;
	uint8_t bits;
	uint8_t parity = 0;

	GPIO_InitStructure.GPIO_Pin = clock;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	tmp1 = tenustimer;
	GPIO_ResetBits(GPIOC, clock);
	// pull clock low for >=100us
	while(tenustimer < (tmp1 + 15));

	// Bring data line low
	GPIO_InitStructure.GPIO_Pin = data;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);
	GPIO_ResetBits(GPIOC, data);
	
	// Release clock
	GPIO_InitStructure.GPIO_Pin = clock;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Might need to wait for the clock to go high after
	// releasing it...
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);

#if 0
	// start bit
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);
#endif

	for(bits = 0; bits < 8; bits++) {
		// Wait for device to bring clock low
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);
		if(cmd & (1<<bits)) {
			GPIO_SetBits(GPIOC, data);
			parity++;
		} else {
			GPIO_ResetBits(GPIOC, data);
		}
		// wait out the remainder of the low clock
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);
	}

	// Data has been sent, now send parity
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);
	if((parity & 0x01) == 0) {
		GPIO_SetBits(GPIOC, data);
	} else {
		GPIO_ResetBits(GPIOC, data);
	}
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);

	// Stop bit, release data line
	GPIO_InitStructure.GPIO_Pin = data;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	while(GPIO_ReadInputDataBit(GPIOC, data) == 0);
	while(GPIO_ReadInputDataBit(GPIOC, data) == 1);
	while(GPIO_ReadInputDataBit(GPIOC, data) == 0);
}

uint16_t PollPS2Frame(uint32_t clock, uint32_t data)
{
	uint16_t ps2frame = 0;
	uint8_t bits;

	for(bits = 0; bits < 11; bits++) {
		do {
			while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);
		} while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);

		do {
			while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);
		} while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);

		// read when clock is high
		ps2frame |= GPIO_ReadInputDataBit(GPIOC, data) << bits;
	}

	return ps2frame;
}

void HandleMacCommand()
{
	switch(mackbd.cmd) {
	case 0x16:
		mackbd.state = kMacKbdStateWriteSettle;
		mackbd.delay = 10;
		mackbd.response = 0x03;
		break;
	case 0x14:
		mackbd.state = kMacKbdStateInstantWait;
		mackbd.delay = 1;
		break;
	case 0x36:
		mackbd.state = kMacKbdStateTest;
		mackbd.delay = 10;
		break;
	case 0x10:
	default:
		mackbd.state = kMacKbdStateInquiryWait;
		mackbd.delay = 25000;
		break;
	};

	return;
}

void USART2_SendHexByte(uint8_t c)
{
	const char base16table[16] = { '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };
	while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
	USART_SendData(USART2, base16table[(c >> 4) & 0x0F]);
	while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
	USART_SendData(USART2, base16table[c & 0x0F]);
}

void USART2_SendString(const char *str)
{
	int i;
	for(i = 0; str[i]; i++) {
		while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
		USART_SendData(USART2, str[i]);
	}
}

void USART2_IRQHandler(void)
{
	if (USART_GetITStatus(USART2, USART_IT_RXNE) != RESET) {
		unsigned char c = USART_ReceiveData(USART2);
		switch(c) {
		case 'r':
			mousex = 100;
			break;
		case 'l':
			mousex = -100;
			break;
		case 'd':
			mousey = 100;
			break;
		case 'u':
			mousey = -100;
			break;
		case 's':
			mousex = 0;
			break;
		};
	}
}
